open util/ordering[State]

sig Image{}
one sig Null in Image{}

sig File {
	loadImage: Image
}
{ loadImage != Null }

sig ImageFile {}

sig State {
	imageFiles: set ImageFile,
	files: imageFiles -> one File,
	images: imageFiles -> one Image
}

pred newImageFile [s, s': State, f: File, i: ImageFile] {
	s'.imageFiles = s.imageFiles + i
	s'.files = s.files ++ i-> f
	s'.images = s.images ++ i -> Null // load the image lazily
}

pred getImage [s, s': State, i: ImageFile, img: Image] {
	s'.imageFiles = s.imageFiles
	s'.files = s.files
	(s.images[i] = Null) => (s'.images = s.images ++ i->s.files[i].loadImage) //load the image
	(s.images[i] != Null) => (s'.images = s.images) //do nothing
	img = s'.images[i]
}

pred init[s: State] { no s.imageFiles and no s.files and no s.images }

fact traces {
	init[first] &&
	all s: State - last | 
		(some f: File, i: ImageFile - s.imageFiles | newImageFile[s, s.next, f, i]) or
		(some i: s.imageFiles, img: Image | getImage[s, s.next, i, img])
}

pred inv[s: State] {
	all i: s.imageFiles, s': State, img: Image | getImage[s,s',i,img] => img != Null
}

pred show {}

assert invHolds {
	all s: State | inv[s]
}

assert invTemp {
	all s1, s2: State |
		lt[s1,s2] => (all i: s1.imageFiles, img1, img2: Image, s1', s2': State | 
			(getImage[s1,s1',i,img1] and getImage[s2,s2',i,img2]) => img1 = img2)
}

run show for 5 but 3 State
check invHolds for 5 but 3 State
check invTemp for 5 but 3 State
